@@ -1,4 +1,4 @@
-# -*- test-case-name: twisted.test.test_defer,twisted.test.test_defgen,twisted.internet.test.test_inlinecb -*-
+# -*- test-case-name: twisted.test.test_defer -*-
 # Copyright (c) Twisted Matrix Laboratories.
 # See LICENSE for details.
 
@@ -18,6 +18,7 @@ Maintainer: Glyph Lefkowitz
 
 from __future__ import division, absolute_import, print_function
 
+import attr
 import traceback
 import types
 import warnings
@@ -731,6 +732,7 @@ class Deferred:
         return self
 
 
+    @failure._extraneous
     def send(self, value=None):
         if self.paused:
             # If we're paused, we have no result to give
@@ -743,6 +745,7 @@ class Deferred:
             # Clear the failure on debugInfo so it doesn't raise "unhandled
             # exception"
             self._debugInfo.failResult = None
+            result.value.__failure__ = result
             raise result.value
         else:
             raise StopIteration(result)
@@ -753,6 +756,91 @@ class Deferred:
     __next__ = send
 
 
+    def asFuture(self, loop):
+        """
+        Adapt a L{Deferred} into a L{asyncio.Future} which is bound to C{loop}.
+
+        @note: converting a L{Deferred} to an L{asyncio.Future} consumes both
+            its result and its errors, so this method implicitly converts
+            C{self} into a L{Deferred} firing with L{None}, regardless of what
+            its result previously would have been.
+
+        @since: Twisted 17.5.0
+
+        @param loop: The asyncio event loop to bind the L{asyncio.Future} to.
+        @type loop: L{asyncio.AbstractEventLoop} or similar
+
+        @param deferred: The Deferred to adapt.
+        @type deferred: L{Deferred}
+
+        @return: A Future which will fire when the Deferred fires.
+        @rtype: L{asyncio.Future}
+        """
+        try:
+            createFuture = loop.create_future
+        except AttributeError:
+            from asyncio import Future
+            def createFuture():
+                return Future(loop=loop)
+        future = createFuture()
+        def checkCancel(futureAgain):
+            if futureAgain.cancelled():
+                self.cancel()
+        def maybeFail(failure):
+            if not future.cancelled():
+                future.set_exception(failure.value)
+        def maybeSucceed(result):
+            if not future.cancelled():
+                future.set_result(result)
+        self.addCallbacks(maybeSucceed, maybeFail)
+        future.add_done_callback(checkCancel)
+        return future
+
+
+    @classmethod
+    def fromFuture(cls, future):
+        """
+        Adapt an L{asyncio.Future} to a L{Deferred}.
+
+        @note: This creates a L{Deferred} from a L{asyncio.Future}, I{not} from
+            a C{coroutine}; in other words, you will need to call
+            L{asyncio.ensure_future},
+            L{asyncio.loop.create_task} or create an
+            L{asyncio.Task} yourself to get from a C{coroutine} to a
+            L{asyncio.Future} if what you have is an awaitable coroutine and
+            not a L{asyncio.Future}.  (The length of this list of techniques is
+            exactly why we have left it to the caller!)
+
+        @since: Twisted 17.5.0
+
+        @param future: The Future to adapt.
+        @type future: L{asyncio.Future}
+
+        @return: A Deferred which will fire when the Future fires.
+        @rtype: L{Deferred}
+        """
+        def adapt(result):
+            try:
+                extracted = result.result()
+            except:
+                extracted = failure.Failure()
+            adapt.actual.callback(extracted)
+        futureCancel = object()
+        def cancel(reself):
+            future.cancel()
+            reself.callback(futureCancel)
+        self = cls(cancel)
+        adapt.actual = self
+        def uncancel(result):
+            if result is futureCancel:
+                adapt.actual = Deferred()
+                return adapt.actual
+            return result
+        self.addCallback(uncancel)
+        future.add_done_callback(adapt)
+        return self
+
+
 
 def _cancelledToTimedOutError(value, timeout):
     """
@@ -809,7 +897,7 @@ def ensureDeferred(coro):
         react(main)
 
     @param coro: The coroutine object to schedule, or a L{Deferred}.
-    @type coro: A Python 3.5+ C{async def} C{coroutine}, a Python 3.3+
+    @type coro: A Python 3.5+ C{async def} C{coroutine}, a Python 3.4+
         C{yield from} using L{types.GeneratorType}, or a L{Deferred}.
 
     @rtype: L{Deferred}
@@ -820,11 +908,7 @@ def ensureDeferred(coro):
         from asyncio import iscoroutine
 
         if iscoroutine(coro) or isinstance(coro, GeneratorType):
-            return _inlineCallbacks(None, coro, Deferred())
-
-    elif version_info >= (3, 3, 0):
-        if isinstance(coro, GeneratorType):
-            return _inlineCallbacks(None, coro, Deferred())
+            return _cancellableInlineCallbacks(coro)
 
     if not isinstance(coro, Deferred):
         raise ValueError("%r is not a coroutine or a Deferred" % (coro,))
@@ -1279,9 +1363,42 @@ def returnValue(val):
 
 
 
-def _inlineCallbacks(result, g, deferred):
+@attr.s
+class _CancellationStatus(object):
     """
-    See L{inlineCallbacks}.
+    Cancellation status of an L{inlineCallbacks} invocation.
+
+    @ivar waitingOn: the L{Deferred} being waited upon (which
+        L{_inlineCallbacks} must fill out before returning)
+
+    @ivar deferred: the L{Deferred} to callback or errback when the generator
+        invocation has finished.
+    """
+
+    deferred = attr.ib()
+    waitingOn = attr.ib(default=None)
+
+
+
+@failure._extraneous
+def _inlineCallbacks(result, g, status):
+    """
+    Carry out the work of L{inlineCallbacks}.
+
+    Iterate the generator produced by an C{@}L{inlineCallbacks}-decorated
+    function, C{g}, C{send()}ing it the results of each value C{yield}ed by
+    that generator, until a L{Deferred} is yielded, at which point a callback
+    is added to that L{Deferred} to call this function again.
+
+    @param result: The last result seen by this generator.  Note that this is
+        never a L{Deferred} - by the time this function is invoked, the
+        L{Deferred} has been called back and this will be a particular result
+        at a point in its callback chain.
+
+    @param g: a generator object returned by calling a function or method
+        decorated with C{@}L{inlineCallbacks}
+
+    @param status: a L{_CancellationStatus} tracking the current status of C{g}
     """
     # This function is complicated by the need to prevent unbounded recursion
     # arising from repeatedly yielding immediately ready deferreds.  This while
@@ -1301,8 +1418,8 @@ def _inlineCallbacks(result, g, deferred):
                 result = g.send(result)
         except StopIteration as e:
             # fell off the end, or "return" statement
-            deferred.callback(getattr(e, "value", None))
-            return deferred
+            status.deferred.callback(getattr(e, "value", None))
+            return
         except _DefGen_Return as e:
             # returnValue() was called; time to give a result to the original
             # Deferred.  First though, let's try to identify the potentially
@@ -1341,11 +1458,11 @@ def _inlineCallbacks(result, g, deferred):
                         ultimateTrace.tb_frame.f_code.co_name,
                         appCodeTrace.tb_frame.f_code.co_name),
                     DeprecationWarning, filename, lineno)
-            deferred.callback(e.value)
-            return deferred
+            status.deferred.callback(e.value)
+            return
         except:
-            deferred.errback()
-            return deferred
+            status.deferred.errback()
+            return
 
         if isinstance(result, Deferred):
             # a deferred was yielded, get the result.
@@ -1354,14 +1471,16 @@ def _inlineCallbacks(result, g, deferred):
                     waiting[0] = False
                     waiting[1] = r
                 else:
-                    _inlineCallbacks(r, g, deferred)
+                    # We are not waiting for deferred result any more
+                    _inlineCallbacks(r, g, status)
 
             result.addBoth(gotResult)
             if waiting[0]:
                 # Haven't called back yet, set flag so that we get reinvoked
                 # and return from the loop
                 waiting[0] = False
-                return deferred
+                status.waitingOn = result
+                return
 
             result = waiting[1]
             # Reset waiting to initial values for next loop.  gotResult uses
@@ -1369,18 +1488,60 @@ def _inlineCallbacks(result, g, deferred):
             # executed once, and if it hasn't been executed yet, the return
             # branch above would have been taken.
 
-
             waiting[0] = True
             waiting[1] = None
 
 
+
+def _cancellableInlineCallbacks(g):
+    """
+    Make an C{@}L{inlineCallbacks} cancellable.
+
+    @param g: a generator object returned by calling a function or method
+        decorated with C{@}L{inlineCallbacks}
+
+    @return: L{Deferred} for the C{@}L{inlineCallbacks} that is cancellable.
+    """
+    def cancel(it):
+        it.callbacks, tmp = [], it.callbacks
+        it.addErrback(handleCancel)
+        it.callbacks.extend(tmp)
+        it.errback(_InternalInlineCallbacksCancelledError())
+    deferred = Deferred(cancel)
+    status = _CancellationStatus(deferred)
+    def handleCancel(result):
+        """
+        Propagate the cancellation of an C{@}L{inlineCallbacks} to the
+        L{Deferred} it is waiting on.
+
+        @param result: An L{_InternalInlineCallbacksCancelledError} from
+            C{cancel()}.
+        @return: A new L{Deferred} that the C{@}L{inlineCallback} generator
+            can callback or errback through.
+        """
+        result.trap(_InternalInlineCallbacksCancelledError)
+        status.deferred = Deferred(cancel)
+        # We would only end up here if the inlineCallback is waiting on
+        # another Deferred.  It needs to be cancelled.
+        awaited = status.waitingOn
+        awaited.cancel()
+        return status.deferred
+    _inlineCallbacks(None, g, status)
     return deferred
 
 
 
+class _InternalInlineCallbacksCancelledError(Exception):
+    """
+    A unique exception used only in L{_cancellableInlineCallbacks} to verify
+    that an L{inlineCallbacks} is being cancelled as expected.
+    """
+
+
+
 def inlineCallbacks(f):
     """
-    inlineCallbacks helps you write L{Deferred}-using code that looks like a
+    L{inlineCallbacks} helps you write L{Deferred}-using code that looks like a
     regular sequential function. For example::
 
         @inlineCallbacks
@@ -1424,13 +1585,18 @@ def inlineCallbacks(f):
                 # will trigger an errback
                 raise Exception('DESTROY ALL LIFE')
 
-    If you are using Python 3.3 or later, it is possible to use the C{return}
-    statement instead of L{returnValue}::
+    It is possible to use the C{return} statement instead of L{returnValue}::
 
         @inlineCallbacks
         def loadData(url):
             response = yield makeRequest(url)
             return json.loads(response)
+
+    You can cancel the L{Deferred} returned from your L{inlineCallbacks}
+    generator before it is fired by your generator completing (either by
+    reaching its end, a C{return} statement, or by calling L{returnValue}).
+    A C{CancelledError} will be raised from the C{yield}ed L{Deferred} that
+    has been cancelled if that C{Deferred} does not otherwise suppress it.
     """
     @wraps(f)
     def unwindGenerator(*args, **kwargs):
@@ -1444,7 +1610,7 @@ def inlineCallbacks(f):
             raise TypeError(
                 "inlineCallbacks requires %r to produce a generator; "
                 "instead got %r" % (f, gen))
-        return _inlineCallbacks(None, gen, Deferred())
+        return _cancellableInlineCallbacks(gen)
     return unwindGenerator
 
 
@@ -1492,6 +1658,20 @@ class _ConcurrencyPrimitive(object):
         return d
 
 
+    def __aenter__(self):
+        """
+        We can be used as an asynchronous context manager.
+        """
+        return self.acquire()
+
+
+    def __aexit__(self, exc_type, exc_val, exc_tb):
+        self.release()
+        # We return False to indicate that we have not consumed the
+        # exception, if any.
+        return succeed(False)
+
+
 
 class DeferredLock(_ConcurrencyPrimitive):
     """
@@ -1573,6 +1753,10 @@ class DeferredSemaphore(_ConcurrencyPrimitive):
     """
 
     def __init__(self, tokens):
+        """
+        @param tokens: initial value of L{tokens} and L{limit}
+        @type tokens: L{int}
+        """
         _ConcurrencyPrimitive.__init__(self)
         if tokens < 1:
             raise ValueError("DeferredSemaphore requires tokens >= 1")
@@ -1822,6 +2006,7 @@ class DeferredFilesystemLock(lockfile.FilesystemLock):
         return d
 
 
+
 __all__ = ["Deferred", "DeferredList", "succeed", "fail", "FAILURE", "SUCCESS",
            "AlreadyCalledError", "TimeoutError", "gatherResults",
            "maybeDeferred", "ensureDeferred",
@@ -1829,4 +2014,6 @@ __all__ = ["Deferred", "DeferredList", "succeed", "fail", "FAILURE", "SUCCESS",
            "returnValue",
            "DeferredLock", "DeferredSemaphore", "DeferredQueue",
            "DeferredFilesystemLock", "AlreadyTryingToLockError",
+           "CancelledError",
           ]
+
